#include <lektor/common.h>
#include <lektor/database.h>
#include <lektor/internal/dbmacro.h>

bool
database_plt_touch(lkt_db *db, const char *name)
{
    static const char *SQL_STMT = "SELECT name FROM playlist WHERE name = ?";
    static const char *SQL_TOUCH =
        "UPDATE playlist SET last_update = strftime('%s', 'now') WHERE name = ?";
    sqlite3_stmt *stmt = NULL;
    bool sta           = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, name, error);

    if (sqlite3_step(stmt) != SQLITE_ROW)
        sta = database_plt_create(db, name);

    else {
        sqlite3_finalize(stmt);
        stmt = NULL;
        SQLITE_PREPARE(db, stmt, SQL_TOUCH, error);
        SQLITE_BIND_TEXT(db, stmt, 1, name, error);
        SQLITE_STEP_DONE(db, stmt, error);
        sta = true;
    }

error:
    if (stmt)
        sqlite3_finalize(stmt);
    return sta;
}

bool
database_plt_create(lkt_db *db, const char *name)
{
    static const char *SQL_STMT = "INSERT INTO playlist (name, last_update) "
                                  "VALUES (?, strftime('%s', 'now'));";
    sqlite3_stmt *stmt          = NULL;
    bool sta                    = false;
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, name, error);
    SQLITE_STEP_DONE(db, stmt, error);
    sta = true;
error:
    sqlite3_finalize(stmt);
    return sta;
}

bool
database_plt_remove(lkt_db *db, const char *name)
{
    static const char *SQL_STMT = "DELETE FROM playlist WHERE name = ?;";
    sqlite3_stmt *stmt          = NULL;
    bool sta                    = false;
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, name, error);
    SQLITE_STEP_DONE(db, stmt, error);
    sta = true;
error:
    sqlite3_finalize(stmt);
    return sta;
}

bool
database_plt_remove_pos(lkt_db *db, const char *name, int pos)
{
    static const char *SQL_STMT =
        "DELETE FROM kara_playlist "
        "WHERE playlist_id = (SELECT id FROM playlist WHERE name = ? LIMIT 1) "
        "AND kara_id = ?;";
    sqlite3_stmt *stmt = NULL;
    bool sta           = false;
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, name, error);
    SQLITE_BIND_INT(db, stmt, 2, pos, error);
    SQLITE_STEP_DONE(db, stmt, error);
    sta = true;
error:
    sqlite3_finalize(stmt);
    return sta;
}

bool
database_plt_clear(lkt_db *db, const char *name)
{
    static const char *SQL_STMT = "DELETE FROM kara_playlist "
                                  "WHERE playlist_id = (SELECT id FROM playlist WHERE name = ?);";
    sqlite3_stmt *stmt          = NULL;
    bool sta                    = false;
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, name, error);
    SQLITE_STEP_DONE(db, stmt, error);
    sta = true;
error:
    sqlite3_finalize(stmt);
    return sta;
}

bool
database_plt_rename(lkt_db *db, const char *old_name, const char *new_name)
{
    static const char *SQL_STMT = "UPDATE playlist SET name = ? WHERE name = ?;";
    sqlite3_stmt *stmt          = NULL;
    bool sta                    = false;
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, new_name, error);
    SQLITE_BIND_TEXT(db, stmt, 2, old_name, error);
    SQLITE_STEP_DONE(db, stmt, error);
    sta = true;
error:
    sqlite3_finalize(stmt);
    return sta;
}

bool
database_plt_export(lkt_db *db, const char *name)
{
    static const char *SQL_SCHEM = "CREATE TABLE IF NOT EXISTS %s.content"
                                   "( kara_id INTEGER PRIMARY KEY NOT NULL CHECK(kara_id > 0) );";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];
    int code, ret = false;
    sqlite3_stmt *stmt = NULL;

    if (___is_sql_str_invalid(name)) {
        LOG_ERROR("DB", "The playlist name '%s' is invalid", name);
        goto error;
    }

    GOTO_UNLESS(strcasecmp(name, LKT_PROTECTED_DATABASE),
                "Name '" LKT_PROTECTED_DATABASE
                "' is protected, you can't use it for a playlist name",
                error);

    safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL_SCHEM, name);
    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    code = sqlite3_step(stmt);

    if (code != SQLITE_DONE && code != SQLITE_OK) {
        LOG_ERROR("DB", "Failed to create schema: %s", sqlite3_errmsg((sqlite3 *)db));
        goto error;
    }

    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

bool
database_plt_import(lkt_db *db, const char *name)
{
    static const char *SQL_SCHEM = "INSERT OR IGNORE INTO kara_playlist (kara_id, playlist_id)"
                                   " SELECT"
                                   "  %s.content.kara_id,"
                                   "  ( SELECT id FROM playlist WHERE name = %s LIMIT 1 )"
                                   " FROM %s.content;\n";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT], ret = false;

    if (___is_sql_str_invalid(name)) {
        LOG_ERROR("DB", "The playlist name '%s' is invalid", name);
        goto error;
    }

    GOTO_UNLESS(strcasecmp(name, LKT_PROTECTED_DATABASE),
                "Name '" LKT_PROTECTED_DATABASE
                "' is protected, you can't use it for a playlist name",
                error);

    safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL_SCHEM, name, name, name);
    SQLITE_EXEC(db, SQL_STMT, error);
    ret = true;
error:
    return ret;
}

bool
database_plt_add_uri(lkt_db *db, const char *name, struct lkt_uri *uri)
{
    bool sta                      = false;
    sqlite3_stmt *stmt            = NULL;
    const LKT_URI_VALUE_TYPE type = lkt_uri_get_value_type(uri);

    if (type == URI_VALUE_TYPE_NULL) {
        LOG_ERROR("DB", "Can't add a NULL URI to the playlist '%s'", name);
        return false;
    }

    if (type == URI_VALUE_TYPE_INTEGER) {
        static const char *SQL =
            "INSERT OR REPLACE INTO kara_playlist (kara_id, playlist_id) "
            " VALUES (?, (SELECT id FROM playlist WHERE name = ? COLLATE NOCASE));";
        SQLITE_PREPARE(db, stmt, SQL, error);
        SQLITE_BIND_INT(db, stmt, 1, (int)lkt_uri_get_value_as_int(uri), error);
        SQLITE_BIND_TEXT(db, stmt, 2, name, error);
        SQLITE_STEP_DONE(db, stmt, error);
        sta = true;
    }

    else if (type == URI_VALUE_TYPE_STRING) {
        static const char *SQL = "INSERT OR REPLACE INTO kara_playlist (kara_id, playlist_id) "
                                 "SELECT"
                                 " kara.id,"
                                 " (SELECT playlist.id FROM playlist WHERE name = ? COLLATE NOCASE)"
                                 "FROM kara WHERE kara.%s LIKE ?;";
        char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];

        safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL, lkt_uri_get_column_name(uri));
        SQLITE_PREPARE(db, stmt, SQL_STMT, error);
        SQLITE_BIND_TEXT(db, stmt, 1, name, error);
        SQLITE_BIND_TEXT(db, stmt, 2, lkt_uri_get_value_as_str(uri), error);
        SQLITE_STEP_DONE(db, stmt, error);
        sta = true;
    }

    else {
        LOG_UNREACHABLE;
    }

error:
    sqlite3_finalize(stmt);
    return sta;
}

bool
database_plt_contais_kara(lkt_db *db, const char *name, int karaid)
{
    sqlite3_stmt *stmt     = NULL;
    static const char *SQL = "SELECT kara_id"
                             " FROM kara_playlist"
                             " JOIN playlist ON playlist.id = playlist_id AND playlist.name = ?"
                             " WHERE kara_id = ?;";

    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_TEXT(db, stmt, 1, name, error);
    SQLITE_BIND_INT(db, stmt, 2, karaid, error);
    SQLITE_STEP_ROW(db, stmt, error);

    sqlite3_finalize(stmt);
    return true;

error:
    if (stmt)
        sqlite3_finalize(stmt);
    return false;
}
